use clichess;
use clichess::{ClientRequest, GameInfo, Player, RecvPositionError, UserRole};
use shakmaty::{fen, Chess, Setup};
use std::collections::HashMap;
use std::fs;
use std::io;
use std::os::unix::net::{UnixListener, UnixStream};
use std::sync::mpsc::{channel, Receiver, Sender, TryRecvError};
use std::sync::{Arc, Condvar, Mutex};

use std::thread;

struct Server {
    id: usize,
    chess_position: Arc<Mutex<Chess>>,
    players: Arc<Mutex<HashMap<usize, (Player, Sender<GameInfo>, Arc<(Mutex<bool>, Condvar)>)>>>,
    others_serv_msg_recv: Receiver<GameInfo>,
    client_message_recv: Receiver<String>,
    cvar: Arc<(Mutex<bool>, Condvar)>,
    opponent_name: Arc<Mutex<Option<String>>>,
    last_move: Arc<Mutex<Option<String>>>,
}

fn main() {
    let version = env!("CARGO_PKG_VERSION");
    println!("Running clichess version {}", version);
    let chess = Arc::new(Mutex::new(Chess::default()));
    let players = Arc::new(Mutex::new(HashMap::new()));
    let mut counter = 0;
    fs::remove_file("/tmp/clichess.socket");

    let listener = UnixListener::bind("/tmp/clichess.socket").unwrap();

    // accept connections and process them, spawning a new thread for each one
    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                let condvar_pair = Arc::new((Mutex::new(false), Condvar::new()));
                let client_message_recv =
                    setup_client_message_recv(&stream, condvar_pair.clone()).unwrap();
                let (others_serv_msg_sender, others_serv_msg_recv) = channel();
                let mut server = Server {
                    id: counter,
                    chess_position: chess.clone(),
                    players: players.clone(),
                    others_serv_msg_recv,
                    client_message_recv,
                    cvar: condvar_pair,
                    opponent_name: Arc::new(Mutex::new(Option::None)),
                    last_move: Arc::new(Mutex::new(Option::None)),
                };
                /* connection succeeded */
                thread::spawn(move || handle_player(stream, &mut server, others_serv_msg_sender));
                counter += 1;
            }
            Err(err) => {
                println!("Could not create a connection: {}", err);
                break;
            }
        }
    }
}

fn handle_player(
    mut stream: UnixStream,
    server: &mut Server,
    others_serv_msg_sender: Sender<GameInfo>,
) {
    loop {
        //wait for input
        let input = server.client_message_recv.recv().unwrap();
        println!("got input: {}", input);
        //route to right function
        match serde_json::from_str(&input) {
            Ok(ClientRequest::FetchAvailableRoles) => fetch_available_roles(&mut stream, server),
            Ok(ClientRequest::Login {
                username,
                pubkey,
                role,
            }) => login(
                &mut stream,
                server,
                username,
                pubkey,
                role,
                others_serv_msg_sender.clone(),
            ),
            Ok(ClientRequest::GetGameInfo) => send_game_info(&mut stream, server),
            Ok(ClientRequest::WaitForNextMove) => {
                //return true if user exited.
                if wait_for_next_move(&mut stream, server) {
                    send_ok(&mut stream);
                    break;
                }
            }
            Ok(ClientRequest::Play(movestr)) => play(&mut stream, server, movestr),
            Ok(ClientRequest::Exit) => {
                send_ok(&mut stream);
                break;
            }
            Err(e) => {
                println!("Bad request: {}", e);
                break;
            }
        }
    }
    player_disconnected(&server);
}

fn login(
    stream: &mut UnixStream,
    server: &Server,
    username: String,
    pubkey: String,
    role: UserRole,
    others_serv_msg_sender: Sender<GameInfo>,
) {
    println!("server {}, player {} is logging in..", server.id, username);
    //check_role
    let mut available_roles_after_choice =
        vec![UserRole::White, UserRole::Black, UserRole::Spectator];
    {
        let players = server.players.lock().unwrap();
        for (_, (player, _, _)) in players.iter() {
            match available_roles_after_choice
                .iter()
                .position(|r| *r == player.role)
            {
                Some(index) => available_roles_after_choice.remove(index),
                None => continue,
            };
        }
    }
    if available_roles_after_choice.contains(&role) {
        create_player(server, username, pubkey, role, others_serv_msg_sender);
        send_ok(stream);
    } else {
        send_ko(stream);
    }
}

fn fetch_available_roles(stream: &mut UnixStream, server: &Server) {
    println!(
        "server {}, sending available role to the player...",
        server.id
    );
    let available_roles = compute_available_roles_to_str(server);
    println!("Computed available_roles as str: {}", available_roles);
    clichess::write_to_stream(stream, available_roles.clone()).unwrap();
}

fn send_game_info(stream: &mut UnixStream, server: &Server) {
    println!("server {}, sending game info...", server.id);
    let chess = server.chess_position.lock().unwrap();
    let opponent_name = server
        .opponent_name
        .lock()
        .unwrap()
        .clone()
        .unwrap_or_default();
    let last_move = server.last_move.lock().unwrap().clone().unwrap_or_default();
    let game_info = GameInfo {
        game_fen: fen::fen(&*chess),
        opponent_name,
        last_move,
    };
    clichess::write_to_stream(stream, serde_json::to_string(&game_info).unwrap()).unwrap();
}

fn wait_for_next_move(stream: &mut UnixStream, server: &mut Server) -> bool {
    println!("server {}, waiting for opponent move...", server.id);
    //acknowledge the request to the client
    send_ok(stream);
    //wait either for the other server, or the client exit.
    match wait_for_opponent_move(server) {
        Ok(game_info) => {
            update_from_game_info(server, &game_info);
            let game_info_serialized = serde_json::to_string(&game_info).unwrap();
            //send game_info to client
            clichess::write_to_stream(stream, game_info_serialized).unwrap();
        }
        Err(_) => {
            return true;
        }
    }
    false
}

fn wait_for_opponent_move(server: &Server) -> Result<GameInfo, RecvPositionError> {
    println!("server id: {}, waiting for move to send...", server.id);
    //wait: either we receive next position from other server threads, or we receive
    //"exit" from the client.
    let mut returned_result = Err(RecvPositionError::UserCanceledError);
    let (lock, cvar) = &*server.cvar;
    let mut message_sent = lock.lock().unwrap();
    *message_sent = false;
    while !*message_sent {
        message_sent = cvar.wait(message_sent).unwrap();
    }
    *message_sent = false;
    match server.others_serv_msg_recv.try_recv() {
        Ok(msg) => {
            returned_result = Ok(msg);
        }
        Err(TryRecvError::Disconnected) => {
            println!("Error: other server thread disconnected while sending move.")
        }
        Err(TryRecvError::Empty) => { /*nothing to do*/ }
    }
    match server.client_message_recv.try_recv() {
        Ok(_) => {
            println!("User exited.");
            returned_result = Err(RecvPositionError::UserCanceledError);
        }
        Err(TryRecvError::Disconnected) => println!("Error: client disconnected."),
        Err(TryRecvError::Empty) => { /*nothing to do*/ }
    }
    returned_result
}

fn update_from_game_info(server: &mut Server, game_info: &GameInfo) {
    let mut chess_position = server.chess_position.lock().unwrap();
    let mut opponent_name = server.opponent_name.lock().unwrap();
    let mut last_move = server.last_move.lock().unwrap();
    *chess_position = clichess::parse_position(&game_info.game_fen);
    if game_info.opponent_name == "" {
        *opponent_name = None;
    } else {
        *opponent_name = Some(game_info.opponent_name.clone());
    }
    if game_info.last_move == "" {
        *last_move = None;
    } else {
        *last_move = Some(game_info.last_move.clone());
    }
}

fn play(stream: &mut UnixStream, server: &Server, movestr: String) {
    println!("server {}, user wants to play...", server.id);
    let mut chess = server.chess_position.lock().unwrap();
    let players = server.players.lock().unwrap();
    let player = players.get(&server.id);
    if player.is_none() || !clichess::is_player_turn(&player.unwrap().0, chess.turn()) {
        send_ko(stream);
        return;
    }
    println!("User tried to play {}", movestr.clone());
    match clichess::try_to_play_move(&chess, movestr.clone()) {
        Ok(played_chess) => *chess = played_chess,
        Err(e) => {
            println!("Error: {}", e);
            send_ko(stream);
            return;
        }
    };
    let player_name = player
        .expect("current server is in the server list")
        .0
        .username
        .clone();

    //send play to other servers
    let game_info_to_send = GameInfo {
        game_fen: fen::fen(&*chess),
        opponent_name: player_name,
        last_move: movestr.clone(),
    };
    for (id, (_, others_serv_msg_sender, cvar_pair)) in players.iter() {
        if server.id != *id {
            others_serv_msg_sender
                .send(game_info_to_send.clone())
                .unwrap();
            let (lock, cvar) = &**cvar_pair;
            let mut message_sent = lock.lock().unwrap();
            *message_sent = true;
            cvar.notify_one();
        }
    }
    let opponent_name = server.opponent_name.lock().unwrap();
    let last_move = server.last_move.lock().unwrap();
    //send updated info to client
    clichess::write_to_stream(
        stream,
        serde_json::to_string(&GameInfo {
            game_fen: fen::fen(&*chess),
            opponent_name: opponent_name.clone().unwrap_or_default(),
            last_move: last_move.clone().unwrap_or_default(),
        })
        .unwrap(),
    )
    .unwrap();
}

fn send_ok(stream: &mut UnixStream) {
    clichess::write_to_stream(stream, String::from("OK")).unwrap();
}

fn send_ko(stream: &mut UnixStream) {
    clichess::write_to_stream(stream, String::from("KO")).unwrap();
}

fn create_player(
    server: &Server,
    username: String,
    public_key: String,
    role: UserRole,
    others_serv_msg_sender: Sender<GameInfo>,
) {
    let mut players = server.players.lock().unwrap();
    let player = Player {
        role,
        username,
        public_key,
    };
    players.insert(
        server.id,
        (player.clone(), others_serv_msg_sender, server.cvar.clone()),
    );
    println!("Created player {}", server.id);
}

fn compute_available_roles_to_str(server: &Server) -> String {
    let mut available_roles = vec![UserRole::White, UserRole::Black, UserRole::Spectator];
    let players = server.players.lock().unwrap();
    for (_, (player, _, _)) in players.iter() {
        match available_roles.iter().position(|r| *r == player.role) {
            Some(index) => available_roles.remove(index),
            None => continue,
        };
    }
    let available_roles_str: Vec<String> = available_roles
        .iter()
        .map(|r| clichess::role_to_str(r))
        .collect();
    available_roles_str.join(",")
}

fn player_disconnected(server: &Server) {
    let mut players = server.players.lock().unwrap();
    let player_and_receiving_buf = players.get(&server.id);
    if player_and_receiving_buf.is_some() {
        let (player, _, _) = player_and_receiving_buf.expect("is some");
        println!("Player {} disconnected.", player.username);
    }
    players.remove(&server.id);
}

fn setup_client_message_recv(
    stream: &UnixStream,
    cvar_pair: Arc<(Mutex<bool>, Condvar)>,
) -> io::Result<Receiver<String>> {
    let (sender, receiver) = channel();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for client message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            //notify the server if waiting
            let (lock, cvar) = &*cvar_pair;
            let mut message_sent = lock.lock().unwrap();
            *message_sent = true;
            cvar.notify_one();
            let send_result = sender.send(buffer);
            // stop the thread during a disconnection
            if send_result.is_err() {
                break;
            }
        }
    });
    Ok(receiver)
}
